//=============================================================================
//
// BSD 3-Clause License
//
// Copyright © 2013-2015, DAVID 3D Solutions GbR
// All rights reserved.
//
// Redistribution and use in source and binary forms, 
// with or without modification, are permitted provided that 
// the following conditions are met:
//
// - Redistributions of source code must retain the above copyright notice, 
//   this list of conditions and the following disclaimer.
// - Redistributions in binary form must reproduce the above copyright notice, 
//   this list of conditions and the following disclaimer in the documentation 
//   and/or other materials provided with the distribution.
// - Neither the name of DAVID 3D Solutions GbR nor the names of its contributors 
//   may be used to endorse or promote products derived from this software 
//   without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//=============================================================================

#include "davidSDK/StructuredLightScannerD4Lib.h"
#include "davidSDK/ShapeFusionD4Lib.h"
#include <sstream>
#include <fstream>

namespace david { 

//=============================================================================
// Global functions
//=============================================================================

//=============================================================================
// StructuredLightScannerD4Lib::ResultFilteringParamsEx
//=============================================================================

StructuredLightScannerD4Lib::ResultFilteringParamsEx::ResultFilteringParamsEx()
	: qualityCheck(0.5)
	, backgroundRemovalEnabled(false)
	, backgroundRemovalEpsilon(2.f)
	, minContrast(10)
	, minDepth(0.f)
	, maxDepth(std::numeric_limits<float>::max())
	, depthThres(10.f)
	, outlierRemoval(0)
{
}


//=============================================================================
// StructuredLightScannerD4Lib::AcquisitionParams
//=============================================================================

StructuredLightScannerD4Lib::AcquisitionParams::AcquisitionParams()
	: minContentChange(20)
	, skipCount(1)
	, maxWaitTime(1.0)
{
}


//=============================================================================
// StructuredLightScannerD4Lib
//=============================================================================

StructuredLightScannerD4Lib::StructuredLightScannerD4Lib(IShapeFusion* shapeFusion)
	: m_cam(d4::cam::newDirectShowCamera())
	, m_camCalib(d4::calib::newCameraCalibration())
	, m_projector(d4::projector::newExtendedDisplayProjector())
	, m_projectorCalib(d4::calib::newProjectorCalibration())
	, m_live(d4::img::newImage())
	, m_sls(d4::sls::newStructuredLightScanner())
	, m_scale(240.0)
	, m_noiseReduction(0)
	, m_shapeFusion(shapeFusion)
	, m_externalCapture(false)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::StructuredLightScannerD4Lib");
	
	m_whiteBalance[0] = 1.0;
	m_whiteBalance[1] = 1.0;
	m_whiteBalance[2] = 1.0;

	AcquisitionParams acq;
	acq.skipCount = 2;
	SetAcquisitionParams(acq);

	SetFrequencies(8);
}


StructuredLightScannerD4Lib::~StructuredLightScannerD4Lib()
{
	
}


void StructuredLightScannerD4Lib::SetResultFilteringParamsEx(const ResultFilteringParamsEx& params)
{
	Guard guard(m_mutex);
	m_resultFilteringParams = params;
}


void StructuredLightScannerD4Lib::SetAcquisitionParams(const AcquisitionParams& params)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SetAcquisitionParams");
	
	m_acqParams = params;
	eec = d4cam_setSkipCount(m_cam.get(), params.skipCount);
}


void StructuredLightScannerD4Lib::SetFrequencies(int frequencies)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SetFrequencies");

	d4sls_CodedLightPhaseShiftParams params;
	eec = d4sls_getCodedLightPhaseShiftParams(m_sls.get(), &params);
	params.frequencies = frequencies;
	eec = d4sls_setCodedLightPhaseShiftParams(m_sls.get(), &params);
	DisplaySetupImage();
}


void StructuredLightScannerD4Lib::SetScreenID(int screenID)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SetScreenID");
	if (screenID >= 0)
	{
		eec = d4projector_openByScreenID(m_projector.get(), screenID);
		DisplaySetupImage();
	}
	else 
	{
		if (m_projector) eec = d4projector_close(m_projector.get());
	}
}


std::vector<std::string> StructuredLightScannerD4Lib::GetAvailableCameraNames()
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::GetAvailableCameraNames");

	int camCount = 0;
	eec = d4cam_getAvailableCamerasCount(m_cam.get(), &camCount);
	std::vector<std::string> names;
	for (int i=0; i < camCount; ++i)
	{
		char name[256];
		eec = d4cam_getAvailableCamerasName(m_cam.get(), name, sizeof(name), i);
		names.push_back(name);
	}
	
	return names;
}


void StructuredLightScannerD4Lib::SelectCamera(const std::string& cameraName, const ImageFormat& imageFormat)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SelectCamera");

	if (cameraName.empty())
	{
		// close camera
		eec = d4cam_close(m_cam.get());
	}
	else
	{
		// open desired camera
		eec = d4cam_open(m_cam.get(), cameraName.c_str(), imageFormat.width, imageFormat.height, imageFormat.pixelFormat, imageFormat.fps);
	}

	m_scan.reset();
	m_texture.reset();
	m_live.reset();
	m_sequence.clear();
	m_externalCapture = false;
}


bool StructuredLightScannerD4Lib::CheckCamera()
{
	Guard guard(m_mutex);
	int result = d4cam_check(m_cam.get());
	return result == d4ok;
}


void StructuredLightScannerD4Lib::SetCameraProperty(const std::string& propertyName, double value, enum CamPropsUsage camPropsUsage)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SetCameraProperty");

	if (camPropsUsage == CamPropsForAll)
	{
		// delete all entries for 'propertyName'.
		m_camPropsForScanning.erase(propertyName);
		m_camPropsForTexturing.erase(propertyName);
	}
	else if (camPropsUsage == CamPropsForScanning)
	{
		m_camPropsForScanning[propertyName] = value;
	}
	else if (camPropsUsage == CamPropsForTexturing)
	{
		m_camPropsForTexturing[propertyName] = value;
	}

	eec = d4cam_setProperty(m_cam.get(), value, propertyName.c_str());
}


double StructuredLightScannerD4Lib::GetCameraProperty(const std::string& propertyName, enum CamPropsUsage camPropsUsage)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::GetCameraProperty");
	
	if (camPropsUsage == CamPropsForAll)
	{
		ThrowException(david::Error_InvalidArgument, "CamPropsForAll is not allowed.", "StructuredLightScannerD4Lib::GetCameraProperty");
	}
	
	if (camPropsUsage == CamPropsForTexturing)
	{
		if (m_camPropsForTexturing.find(propertyName) != m_camPropsForTexturing.end())
		{
			return m_camPropsForTexturing[propertyName];
		}
	}
	else
	{
		if (m_camPropsForScanning.find(propertyName) != m_camPropsForScanning.end())
		{
			return m_camPropsForScanning[propertyName];
		}
	}

	double value = 0.0;
	
	eec = d4cam_getProperty(m_cam.get(), &value, propertyName.c_str());
	return value;
}


std::vector<std::string> StructuredLightScannerD4Lib::GetCameraPropertyNames(bool onlyAvailable)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::GetCameraPropertyNames");

	int count = 0;
	eec = d4cam_getPropertyCount(m_cam.get(), &count, onlyAvailable);
	std::vector<std::string> names;
	for (int i=0; i < count; ++i)
	{
		char name[256];
		eec = d4cam_getPropertyName(m_cam.get(), name, sizeof(name), i, onlyAvailable);
		names.push_back(name);
	}

	return names;
}


void StructuredLightScannerD4Lib::SelectImageSequence(const std::string& directory)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SelectImageSequence");
	
	// TODO
	eec = d4error_NotImplemented;
}


int StructuredLightScannerD4Lib::GetPatternCount()
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::GetPatternCount");
	
	int count = 0;
	eec = d4sls_getPatternCount(m_sls.get(), &count);
	return count;
}


void StructuredLightScannerD4Lib::GetPatternImage(uint8_t* data, int width, int height, PatternPixelFormat pixelFormat, int patternIndex)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::GetPatternImage");
	if (data == nullptr)
	{
		return ThrowException(david::Error_InvalidArgument, "data is nullptr.", eec.GetSource());
	}

	auto img = d4::img::newImage();
	eec = d4sls_getPatternImage(m_sls.get(), img.get(), width, height, pixelFormat, patternIndex);
	eec = d4img_getImageData(img.get(), data, GetImageSize(width, height, (PixelFormat)pixelFormat), pixelFormat);
}


void StructuredLightScannerD4Lib::SetCapturedImage(const uint8_t* data, int width, int height, CapturedPixelFormat pixelFormat, int patternIndex)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SetCapturedImage");

	// Close camera if open.
	if (d4cam_check(m_cam.get()) == d4ok) d4cam_close(m_cam.get());

	eec("setup sequence");
	int patternCount = 0;
	eec = d4sls_getPatternCount(m_sls.get(), &patternCount);
	if (patternIndex < 0 || patternIndex >= patternCount)
	{
		return ThrowException(Error_InvalidArgument, "patternIndex is invalid.", eec.GetSource());
	}
	if (patternCount != int(m_sequence.size()))
	{
		m_sequence.resize(patternCount);
		for (int i=0; i < patternCount; ++i)
		{
			m_sequence[i] = d4::img::newImage();
		}
	}

	eec("set captured image");
	auto img = m_sequence[patternIndex];
	eec = d4img_setImageFormat(img.get(), width, height, pixelFormat);
	eec = d4img_setImageData(img.get(), data, GetImageSize(width, height, (PixelFormat)pixelFormat), pixelFormat);

	m_externalCapture = true;
}


void StructuredLightScannerD4Lib::SetTextureImage(const uint8_t* data, int width, int height, PixelFormat pixelFormat)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SetTextureImage");

	if (data == nullptr)
	{
		m_texture.reset();
		return;
	}

	if (m_texture == nullptr)
	{
		m_texture = d4::img::newImage();
	}

	eec = d4img_setImageFormat(m_texture.get(), width, height, PixelFormat::RGB24);
	eec = d4img_setImageData(m_texture.get(), data, GetImageSize(width, height, pixelFormat), pixelFormat);
}


void StructuredLightScannerD4Lib::GetLiveImage(std::vector<uint8_t>& pixelValues, int& width, int& height)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::GetLiveImage");

	eec = d4cam_getImageFormat(m_cam.get(), &width, &height, NULL, NULL);
	pixelValues.resize(width*height);
	if (pixelValues.size() > 0)
	{
		eec = d4cam_captureImageAsync(m_cam.get(), 0, 1.0, 0);
		eec = d4cam_getCapturedImage(m_cam.get(), m_live.get(), d4img_Y800);
		eec = d4img_getImageData(m_live.get(), pixelValues.data(), pixelValues.size(), d4img_Y800);
	}
}


void StructuredLightScannerD4Lib::SetCodedLightPhaseShiftMode(const CodedLightPhaseShiftParams& params)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec;

	// Get.
	d4sls_CodedLightPhaseShiftParams slsParams;
	eec = d4sls_getCodedLightPhaseShiftParams(m_sls.get(), &slsParams);
	
	// Set.
	if (params.frequencies && params.frequencies.value()!=0) slsParams.frequencies = params.frequencies.value();  // todo: AutoFrequencies
	if (params.color) slsParams.colorSelect = params.color.value();
	if (params.brightness) slsParams.brightness = params.brightness.value();
	if (params.inverse) slsParams.inverse = params.inverse.value();
	if (params.orientation) slsParams.orientation = params.orientation.value();
	if (params.noiseReduction) m_noiseReduction = params.noiseReduction.value();
	if (params.shifts) slsParams.shifts = params.shifts.value();

	eec = d4sls_setCodedLightPhaseShiftParams(m_sls.get(), &slsParams);
}


CodedLightPhaseShiftParams StructuredLightScannerD4Lib::GetCodedLightPhaseShiftParams()
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::GetCodedLightPhaseShiftParams");

	// Get.
	d4sls_CodedLightPhaseShiftParams slsParams;
	eec = d4sls_getCodedLightPhaseShiftParams(m_sls.get(), &slsParams);

	CodedLightPhaseShiftParams params;
	params.frequencies = slsParams.frequencies;
	params.color = (PatternColor) slsParams.colorSelect;
	params.brightness = slsParams.brightness;
	params.inverse = slsParams.inverse;
	params.noiseReduction = m_noiseReduction;
	params.orientation = (Orientation) slsParams.orientation;
	params.shifts = slsParams.shifts;
	return params;
}


void StructuredLightScannerD4Lib::SetResultFiltering(const ResultFilteringParams& params)
{
	Guard guard(m_mutex);
	if (params.backgroundRemovalEnabled) m_resultFilteringParams.backgroundRemovalEnabled = params.backgroundRemovalEnabled.value();
	if (params.backgroundRemovalEpsilon) m_resultFilteringParams.backgroundRemovalEpsilon = (float)params.backgroundRemovalEpsilon.value();
	if (params.qualityCheck) m_resultFilteringParams.qualityCheck = (float)params.qualityCheck.value();
	if (params.outlierRemoval) m_resultFilteringParams.outlierRemoval = params.outlierRemoval.value();
}


ResultFilteringParams StructuredLightScannerD4Lib::GetResultFiltering()
{
	Guard guard(m_mutex);
	ResultFilteringParams params;
	params.backgroundRemovalEnabled = m_resultFilteringParams.backgroundRemovalEnabled;
	params.backgroundRemovalEpsilon = m_resultFilteringParams.backgroundRemovalEpsilon;
	params.qualityCheck = m_resultFilteringParams.qualityCheck;
	return params;
}


void StructuredLightScannerD4Lib::ImportCalibration(const std::string& camCalibFilename, const std::string& projectorCalibFilename)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::ImportCalibration");
	
	eec = d4calib_importCalibration(m_projectorCalib.get(), projectorCalibFilename.c_str());
	eec = d4calib_importCalibration(m_camCalib.get(), camCalibFilename.c_str());
}


void StructuredLightScannerD4Lib::ExportCalibration(const std::string& camCalibFilename, const std::string& projectorCalibFilename)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::ExportCalibration");

	eec = d4calib_exportCalibration(m_projectorCalib.get(), projectorCalibFilename.c_str());
	eec = d4calib_exportCalibration(m_camCalib.get(), camCalibFilename.c_str());
}


void StructuredLightScannerD4Lib::ChangeCameraCalibrationResolution(int newWidth, int newHeight)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::ChangeCameraCalibrationResolution");

	eec = d4calib_changeCalibrationResolution(m_camCalib.get(), newWidth, newHeight);
}


std::vector<CalibPoint> StructuredLightScannerD4Lib::Calibrate(double scale)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::Calibrate");
	
	m_scale = scale;

	// Save old params for later reset.
	d4sls_CodedLightPhaseShiftParams oldParams;
	eec = d4sls_getCodedLightPhaseShiftParams(m_sls.get(), &oldParams);

	eec("capture image for camera calibration");
	auto calibImg = d4::img::newImage();
	{
		eec = d4cam_captureImageAsync(m_cam.get(), m_acqParams.minContentChange, 1.0, 10);
		uint8_t brightness = uint8_t((oldParams.brightness*3)/4);
		eec = d4projector_displayUniformColor(m_projector.get(), brightness, brightness, brightness);
		eec = d4cam_getCapturedImage(m_cam.get(), calibImg.get(), d4img_Y800);
	}

	// We use high quality for SLS calibration.
	// Important: Ensure that old parameters are restored.
	eec("capture sequence for projector calibration");
	d4sls_CodedLightPhaseShiftParams params = oldParams;
	params.inverse = true;
	params.shifts = 15;
	params.orientation = d4sls_BothOrientations;
	eec = d4sls_setCodedLightPhaseShiftParams(m_sls.get(), &params);
	std::vector<d4calib_CalibPoint> calibPoints;
	try
	{
		CaptureSequence();
		std::vector<d4img_ptr> capturedImages;
		{
			// Convert captured images to depth image.
			for (size_t i=0; i < m_sequence.size(); ++i)
			{
				capturedImages.push_back(m_sequence[i].get());
			}
		}
		if (capturedImages.size() < 1) eec = d4error_Fail;

		// calibrate
		eec("calibrate");
		int calibError;
		bool compensatePatternOffsets = true;
		int calibPointCount = 1000;
		calibPoints.resize(calibPointCount);
		eec = d4calib_calibrate(m_camCalib.get(), calibImg.get(), m_projectorCalib.get(), m_sls.get(), &capturedImages[0], (int)capturedImages.size(), m_scale, compensatePatternOffsets, 
			&calibError, &calibPoints[0], &calibPointCount);
		calibPoints.resize(calibPointCount);
	}
	catch (david::Exception&)
	{
		d4sls_setCodedLightPhaseShiftParams(m_sls.get(), &oldParams);
		throw;
	}

	// Restore old pattern parameters.
	eec("restore");
	eec = d4sls_setCodedLightPhaseShiftParams(m_sls.get(), &oldParams);

	// Display chessboard or setup image.
	if (!DisplayChessboardImage()) DisplaySetupImage();

	// Notification about successful calibration.
	OnCalibrationDone(calibPoints);

	// Return calibration error.
	std::vector<CalibPoint> _calibPoints(calibPoints.size());
	for (size_t c=0; c < calibPoints.size(); ++c)
	{
		for (int j=0; j < 3; ++j) _calibPoints[c].worldRefPos3d[j] = calibPoints[c].worldRefPos3d[j];
		for (int j=0; j < 2; ++j)
		{
			_calibPoints[c].projectedPos2d[j] = calibPoints[c].projectedPos2d[j];
			_calibPoints[c].measuredPos2d[j] = calibPoints[c].measuredPos2d[j];
		}
	}
	return _calibPoints;
}


int StructuredLightScannerD4Lib::Scan(bool grabTexture, bool setAsBackground)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::Scan");

	// check if camera and projector are calibrated:
	bool calibrated = false;
	d4calib_isCalibrated(m_projectorCalib.get(), &calibrated);
	if (!calibrated) eec = d4error_NotCalibrated;
	d4calib_isCalibrated(m_camCalib.get(), &calibrated);
	if (!calibrated) eec = d4error_NotCalibrated;

	// capture images
	if (!m_externalCapture)
	{
		eec = d4projector_check(m_projector.get());
		eec = d4cam_check(m_cam.get());

		SwitchCameraProperties(CamPropsForScanning);
	
		eec("capture");
		CaptureSequence();
		if (grabTexture) CaptureTexture(true);
	}

	// Sequence empty? --> Nothing to do
	if (m_sequence.empty()) return -1;

	eec("compute depth image");
	auto depthImage = d4::fimg::newImage();
	auto qualityImage = d4::fimg::newImage();
	{
		// Convert captured images to depth image.
		std::vector<d4img_ptr> capturedImages;
		for (size_t i=0; i < m_sequence.size(); ++i)
		{
			capturedImages.push_back(m_sequence[i].get());
		}
		
		eec = d4sls_computeDepthImage(m_sls.get(), depthImage.get(), qualityImage.get(), 
			&capturedImages[0], int(capturedImages.size()), m_camCalib.get(), m_projectorCalib.get(), m_resultFilteringParams.minContrast);
	}

	eec("set or remove background");
	// Set or remove background.
	if (setAsBackground)
	{
		m_background = depthImage;
	}
	else if (m_background && m_resultFilteringParams.backgroundRemovalEnabled)
	{
		d4scan_removeBackground(depthImage.get(), m_background.get(), float(m_resultFilteringParams.backgroundRemovalEpsilon));
	}

	eec("depth image to mesh");
	m_scan = d4::mesh::newMesh();
	eec = d4scan_convertDepthImageToMesh(m_scan.get(), depthImage.get(), qualityImage.get(), m_camCalib.get(), 
		m_resultFilteringParams.minDepth, m_resultFilteringParams.maxDepth, m_resultFilteringParams.depthThres, m_resultFilteringParams.qualityCheck, m_resultFilteringParams.outlierRemoval);

	eec("set texture");
	if (grabTexture && m_texture)
	{
		eec = d4mesh_setTexture(m_scan.get(), m_texture.get(), 0);
	}
	else
	{
		m_texture.reset();
	}

	int vertexCount = 0;
	eec = d4mesh_getVertexCount(m_scan.get(), &vertexCount, -1);
	
	return vertexCount;
}


void StructuredLightScannerD4Lib::GrabTexture()
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::GrabTexture");
	
	CaptureTexture(true);

	// Set texture.
	if (m_scan && m_texture)
	{
		eec = d4mesh_setTexture(m_scan.get(), m_texture.get(), 0);
	}
}


void StructuredLightScannerD4Lib::AutoWhiteBalance()
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::AutoWhiteBalance");
	
	CaptureTexture(false);

	eec = d4img_computeWhiteBalance(m_texture.get(), &m_whiteBalance[0], &m_whiteBalance[1], &m_whiteBalance[2]);
	eec = d4img_applyWhiteBalance(m_texture.get(), m_whiteBalance[0], m_whiteBalance[1], m_whiteBalance[2]);
}


int StructuredLightScannerD4Lib::AddScanToShapeFusion()
{
	Guard guard(m_mutex);
	ShapeFusionD4Lib* shapeFusion = dynamic_cast<ShapeFusionD4Lib*>(m_shapeFusion);
	if (shapeFusion && m_scan)
	{
		return shapeFusion->AddToList(m_scan);
	}
	else return -1;
}


void StructuredLightScannerD4Lib::ExportMesh(const std::string& filename)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::ExportMesh");

	eec = d4mesh_export(m_scan.get(), filename.c_str());
}


void StructuredLightScannerD4Lib::CaptureSequence()
{
	david::ErrorExceptionConverter eec("StructuredLightScannerD4Lib::CaptureSequence");

	// Ensure that sequence has correct size.
	eec("setup sequence");
	int patternCount = 0;
	eec = d4sls_getPatternCount(m_sls.get(), &patternCount);
	if (patternCount != int(m_sequence.size()))
	{
		m_sequence.resize(patternCount);
		for (int i=0; i < patternCount; ++i)
		{
			m_sequence[i] = d4::img::newImage();
		}
	}

	eec("capture");
	auto patternImage = d4::img::newImage();
	for (int i=0; i < patternCount; ++i)
	{
		int projectorWidth=0, projectorHeight=0, projectorPixelFormat=0;
		eec = d4projector_getImageFormat(m_projector.get(), &projectorWidth, &projectorHeight, &projectorPixelFormat);
		eec = d4cam_captureImageAsync(m_cam.get(), m_acqParams.minContentChange, m_acqParams.maxWaitTime, m_noiseReduction);
		eec = d4sls_getPatternImage(m_sls.get(), patternImage.get(), projectorWidth, projectorHeight, projectorPixelFormat, i); 
		eec = d4projector_displayImage(m_projector.get(), patternImage.get());
		eec = d4cam_getCapturedImage(m_cam.get(), m_sequence[i].get(), d4img_Y800);
	}

	/*
	for (int i=0; i < patternCount; ++i)
	{
		auto temp = d4::img::newImage();
		eec = d4img_convert(temp.get(), d4img_RGB24, m_sequence[i].get());
		{
			std::stringstream ss;
			ss << "M:\\Temp\\003\\";
			if (i+1 < 10) ss << "0";
			ss << (i+1) << ".bmp";
			eec = d4img_export(temp.get(), ss.str().c_str());
		}
	}
	*/

	DisplaySetupImage();
}


void StructuredLightScannerD4Lib::CaptureTexture(bool applyWhiteBalance)
{
	david::ErrorExceptionConverter eec("StructuredLightScannerD4Lib::CaptureTexture");

	SwitchCameraProperties(CamPropsForTexturing);

	eec("new texture");
	m_texture = d4::img::newImage();

	bool isColorCamera = true;
	{
		int pixelFormat;
		eec = d4cam_getImageFormat(m_cam.get(), NULL, NULL, &pixelFormat, NULL);
		if (pixelFormat == d4img_Y800) isColorCamera = false;
	}

	if (isColorCamera)
	{
		eec("capture texture from color cam");
		// Get brightness for texturing.
		d4sls_CodedLightPhaseShiftParams params;
		eec = d4sls_getCodedLightPhaseShiftParams(m_sls.get(), &params);
		uint8_t b = (uint8_t) params.brightness;

		// Capture texture.
		eec = d4cam_captureImageAsync(m_cam.get(), m_acqParams.minContentChange, m_acqParams.maxWaitTime, m_noiseReduction);
		eec = d4projector_displayUniformColor(m_projector.get(), b, b, b); 
		eec = d4cam_getCapturedImage(m_cam.get(), m_texture.get(), d4img_RGB24);
	}
	else
	{
		eec("capture texture from grayscale cam");
		// Create temporary images.
		std::vector<std::shared_ptr<d4img>> images(5);
		for (auto& image : images)
		{
			image = d4::img::newImage();
		}

		// Get brightness for texturing.
		d4sls_CodedLightPhaseShiftParams params;
		eec = d4sls_getCodedLightPhaseShiftParams(m_sls.get(), &params);
		uint8_t b = (uint8_t) params.brightness;

		// Get color components for displaying.
		uint8_t reds[]   = {b, 0, 0, 0, 0};
		uint8_t greens[] = {0, 0, b, 0, 0};
		uint8_t blues[]  = {0, 0, 0, 0, b};

		// Display colors and capture images.
		for (int i=0; i < 5; ++i)
		{
			eec = d4cam_captureImageAsync(m_cam.get(), m_acqParams.minContentChange, m_acqParams.maxWaitTime, m_noiseReduction);
			eec = d4projector_displayUniformColor(m_projector.get(), reds[i], greens[i], blues[i]); 
			eec = d4cam_getCapturedImage(m_cam.get(), images[i].get(), d4img_Y800);
		}

		std::vector<d4img_ptr> nativeImagePtrs(4);
		nativeImagePtrs[0] = images[0].get(); // red
		nativeImagePtrs[1] = images[2].get(); // green
		nativeImagePtrs[2] = images[4].get(); // blue
		nativeImagePtrs[3] = images[1].get(); // black
		eec = d4img_computeColorImage(m_texture.get(), nativeImagePtrs.data(), (int)nativeImagePtrs.size());
	}

	if (applyWhiteBalance)
	{
		eec = d4img_applyWhiteBalance(m_texture.get(), m_whiteBalance[0], m_whiteBalance[1], m_whiteBalance[2]);
	}
}


void StructuredLightScannerD4Lib::SwitchCameraProperties(CamPropsUsage usage)
{
	if (usage == CamPropsForScanning)
	{
		std::map<std::string, double>::iterator it = m_camPropsForScanning.begin();
		for (; it != m_camPropsForScanning.end(); ++it)
		{
			d4cam_setProperty(m_cam.get(), it->second, it->first.c_str());
		}
	}
	else if (usage == CamPropsForTexturing)
	{
		std::map<std::string, double>::iterator it = m_camPropsForTexturing.begin();
		for (; it != m_camPropsForTexturing.end(); ++it)
		{
			d4cam_setProperty(m_cam.get(), it->second, it->first.c_str());
		}
	}
}


void StructuredLightScannerD4Lib::DisplaySetupImage()
{
	david::ErrorExceptionConverter eec("StructuredLightScannerD4Lib::DisplaySetupImage");

	if (d4projector_check(m_projector.get()) != d4ok) return;

	int width, height, pixelFormat;
	eec = d4projector_getImageFormat(m_projector.get(), &width, &height, &pixelFormat);
	auto img = d4::img::newImage();
	eec = d4sls_getSetupImage(m_sls.get(), img.get(), width, height, pixelFormat);
	eec = d4projector_displayImage(m_projector.get(), img.get());
}


bool StructuredLightScannerD4Lib::DisplayChessboardImage()
{
	david::ErrorExceptionConverter eec("StructuredLightScannerD4Lib::DisplayChessboardImage");

	// check if projector is calibrated
	if (d4projector_check(m_projector.get()) != d4ok) return false;
	bool calibrated = false;
	if (d4calib_isCalibrated(m_projectorCalib.get(), &calibrated) != d4ok || !calibrated) return false;

	// get image format of projector
	int width, height, pixelFormat;
	eec = d4projector_getImageFormat(m_projector.get(), &width, &height, &pixelFormat);

	// get chessboard image and display it
	auto img = d4::img::newImage();
	int result = d4sls_getChessboardImage(m_sls.get(), img.get(), width, height, pixelFormat, m_projectorCalib.get(), m_scale);
	if (result != d4ok) return false;
	eec = d4projector_displayImage(m_projector.get(), img.get());
	return true;
}


void StructuredLightScannerD4Lib::OnCalibrationDone(const std::vector<d4calib_CalibPoint>& calibPoints)
{
	/*
	std::ofstream f("calibration_points.txt");
	for (size_t i=0; i < calibPoints.size(); ++i)
	{
		d4calib_CalibPoint cp = calibPoints[i];
		f << "index:            " << i << std::endl;
		f << "  worldRefPos3d:  " << cp.worldRefPos3d[0] << " " << cp.worldRefPos3d[1] << " " << cp.worldRefPos3d[2] << std::endl;
		f << "  measuredPos2d:  " << cp.measuredPos2d[0] << " " << cp.measuredPos2d[1] << std::endl;
		f << "  projectedPos2d: " << cp.projectedPos2d[0] << " " << cp.projectedPos2d[1] << std::endl;
	}
	*/
}


void StructuredLightScannerD4Lib::ExportBackgroundDepthmap(const std::string& filename)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::ExportBackgroundDepthmap");

	if (m_background == NULL) eec = d4error_MissingObject;
	eec = d4fimg_export(m_background.get(), filename.c_str());
}


void StructuredLightScannerD4Lib::ImportBackgroundDepthmap(const std::string& filename)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::ImportBackgroundDepthmap");

	auto backgrnd = d4::fimg::newImage();
	eec = d4fimg_import(backgrnd.get(), filename.c_str());
	m_background = backgrnd;
}


void StructuredLightScannerD4Lib::GetBackgroundDepthmap(std::vector<float>& depthValues, int &width, int &height)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::GetBackgroundDepthmap");

	if (m_background == NULL) eec = d4error_MissingObject;

	eec = d4fimg_getImageFormat(m_background.get(), &width, &height);
	depthValues.resize(width*height);
	eec = d4fimg_getImageData(m_background.get(), &depthValues[0], width*height*sizeof(float));
}


void StructuredLightScannerD4Lib::SetBackgroundDepthmap(const std::vector<float>& depthValues, int width, int height)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SetBackgroundDepthmap");

	if (width*height != depthValues.size()) ThrowException(david::Error_InvalidArgument, "width and height do not fit", "StructuredLightScannerD4Lib::SetBackgroundDepthmap");

	auto background = d4::fimg::newImage();
	eec = d4fimg_setImageFormat(background.get(), width, height);
	eec = d4fimg_setImageData(background.get(), &depthValues[0], width*height*sizeof(float));

	m_background = background;
}


void StructuredLightScannerD4Lib::SetBackgroundPlanes(const std::vector<Plane>& planes)
{
	Guard guard(m_mutex);
	ErrorExceptionConverter eec("StructuredLightScannerD4Lib::SetBackgroundPlanes");

	// Get width and height:
	int camResX=0;
	int camResY=0;
	eec = d4calib_getCalibrationResolution(m_camCalib.get(), &camResX, &camResY);
	if (camResX<=0 || camResY<=0) ThrowException(david::Error_FormatNotAvailable, "Could not determine camera resolution", "StructuredLightScannerD4Lib::SetBackgroundPlanes");

	// Prepare new depth map:
	auto background = d4::fimg::newImage();
	eec = d4fimg_setImageFormat(background.get(), camResX, camResY);

	// Fill depth map:
	eec = d4sls_planesToDepthImage(m_sls.get(), background.get(), m_camCalib.get(), (int)planes.size(), (const d4sls_Plane*)planes.data());

	// store to m_background:
	m_background = background;
}


} // namespace

